import React from 'react';
import { ButtonDropdown, DropdownToggle, DropdownMenu, DropdownItem } from 'reactstrap';
import ReactMarkdown from 'react-markdown';
import {observer} from 'mobx-react';
import FileSaver from 'file-saver';

import Icon from '../components/Icon.js';

import Paragraph from './components/Paragraph.js';
import Blockquote from './components/Blockquote.js';
import Break from './components/Break.js';
import Code from './components/Code.js';
import InlineCode from './components/InlineCode.js';
import Emphasis from './components/Emphasis.js';
import Delete from './components/Delete.js';
import Heading from './components/Heading.js';
import Image from './components/Image.js';
import Table from './components/Table.js';
import Strong from './components/Strong.js';
import List from './components/List.js';
import ListItem from './components/ListItem.js';
import Link from './components/Link.js';

@observer
export default class Editor extends React.Component {

    constructor(props){
        super(props);
        this.state= {
            textareaFocus: false,
            textDropdown: false
        };
    }

    saveFile(){
        let blob = new Blob([this.props.store.content], { type: 'text/markdown;charset=utf-8' });
        FileSaver.saveAs(blob, "file.md");
    }

    // textarea events

    onContentChange(e){
        if(!this.state.textareaFocus){
            return false;
        }
        this.props.store.onContentChange(e.target.value, e.target.selectionStart, e.target.selectionEnd);
    }

    onSelectionChange(e){
        if(!this.state.textareaFocus){
            return false;
        }
        this.props.store.onSelectionChange(e.target.selectionStart, e.target.selectionEnd);
    }

    onFocus(){
        this.setState({
            textareaFocus: true
        });
    }

    onBlur(){
        this.setState({
            textareaFocus: false
        });
    }

    onScroll(e){
        if(this.state.textareaFocus){
            var scrollTop  = e.target.scrollTop  / e.target.scrollTopMax;
            var scrollLeft = e.target.scrollLeft / e.target.scrollLeftMax;
            this.props.store.onScrollChange(scrollTop, scrollLeft);
        } else {
            this.textarea.scrollTop  = this.props.store.scrollTop  * this.textarea.scrollTopMax;
            this.textarea.scrollLeft = this.props.store.scrollLeft * this.textarea.scrollLeftMax;
        }
    }

    // button events
    
    onTextDropdownToggle(){
        this.setState({
            textDropdown: !this.state.textDropdown
        });
    }

    setStrong(){
        this.setInlineBlock(
            'strong',
            ['image', 'imageReference', 'code', 'break', 'thematicBreak'],
            (line, modeAdd) => {
                line = line.replace(/\*\*/g, '');
                if(modeAdd && line.length > 1){
                    line = '**'+line+'**';
                }
                return line;
            }
        );
    }

    setEmphasis(){
        this.setInlineBlock(
            'emphasis',
            ['image', 'imageReference', 'code', 'break', 'thematicBreak'],
            (line, modeAdd) => {
                line = line.replace(/_/g, '');
                if(modeAdd && line.length > 1){
                    line = '_'+line+'_';
                }
                return line;
            }
        );
    }

    setDelete(){
        this.setInlineBlock(
            'delete',
            ['image', 'imageReference', 'code', 'break', 'thematicBreak'],
            (line, modeAdd) => {
                line = line.replace(/~~/g, '');
                if(modeAdd && line.length > 1){
                    line = '~~'+line+'~~';
                }
                return line;
            }
        );
    }

    setLink(){
        this.setInlineBlock(
            'link',
            ['image', 'imageReference', 'code', 'break', 'thematicBreak'],
            (line, modeAdd) => {
                if(modeAdd){
                    line = '['+line+'](url)';
                } else {
                    var desc = line.match(/\[[^\]]*\]/g);
                    if(desc){
                        desc = desc[0].substr(1, desc[0].length - 2);
                    }
                    return desc;
                }
                return line;
            }
        );
    }

    setImage(){
        this.setInlineBlock(
            'image',
            ['code', 'break', 'thematicBreak'],
            (line, modeAdd) => {
                if(modeAdd){
                    line = '!['+line+'](url)';
                } else {
                    var desc = line.match(/\!\[[^\]]*\]/g);
                    if(desc){
                        desc = desc[0].substr(2, desc[0].length - 3);
                    }
                    return desc;
                }
                return line;
            }
        );
    }

    setHeader1(){
        this.setTextType('heading-1', '# ');
    }

    setHeader2(){
        this.setTextType('heading-2', '## ');
    }

    setHeader3(){
        this.setTextType('heading-3', '### ');
    }

    setHeader4(){
        this.setTextType('heading-4', '#### ');
    }

    setHeader5(){
        this.setTextType('heading-5', '##### ');
    }

    setHeader6(){
        this.setTextType('heading-6', '###### ');
    }

    setParagraph(){
        this.setTextType('paragraph', '');
    }

    setTextType(type, add){
        this.setLineBlock(
            {
                type:           type,
                forbiddenTypes: ['break', 'thematicBreak', 'image', 'imageReference', 'table', 'code']
            },
            (lines, modeAdd) => {
                for(var i = 0; i < lines.length; i++){
                    var n = this.normalizeLine(lines[i]);
                    if(!modeAdd){
                        add = '';
                    }
                    lines[i] = n.whitespace + add + n.line;
                }
                return lines;
            }
        );
    }

    setCode(){
        this.setLineBlock(
            {
                type:           'code',
                forbiddenTypes: [],
                completeType:   true
            },
            (lines, modeAdd) => {
                if(modeAdd){
                    lines.unshift('```');
                    lines.push('```');
                } else {
                    if(lines[0].substr(0, 3) === '```'){
                        lines.shift();
                    }
                    if(lines[lines.length - 1].substr(0, 3) === '```'){
                        lines.pop();
                    }
                }
                return lines;
            }
        );
    }

    setMath(){
        this.setLineBlock(
            {
                type:           'math',
                forbiddenTypes: [],
                completeType:   true
            },
            (lines, modeAdd) => {
                if(modeAdd){
                    lines.unshift('```math');
                    lines.push('```');
                } else {
                    if(lines[0].substr(0, 3) === '```'){
                        lines.shift();
                    }
                    if(lines[lines.length - 1].substr(0, 3) === '```'){
                        lines.pop();
                    }
                }
                return lines;
            }
        );
    }

    setPlotly(){
        this.setLineBlock(
            {
                type:           'plotly',
                forbiddenTypes: [],
                completeType:   true
            },
            (lines, modeAdd) => {
                if(modeAdd){
                    lines.unshift('```plotly');
                    lines.push('```');
                } else {
                    if(lines[0].substr(0, 3) === '```'){
                        lines.shift();
                    }
                    if(lines[lines.length - 1].substr(0, 3) === '```'){
                        lines.pop();
                    }
                }
                return lines;
            }
        );
    }

    setBlockquote(){
        this.setLineBlock(
            {
                type:           'blockquote',
                forbiddenTypes: ['break', 'thematicBreak', 'image', 'imageReference', 'table', 'code'],
                completeType:   true
            },
            (lines, modeAdd) => {
                for(var i = 0; i < lines.length; i++){
                    var n = this.normalizeLine(lines[i]);
                    var add = '';
                    if(modeAdd){
                        add = '> ';
                    }
                    lines[i] = n.whitespace + add + n.line;
                }
                return lines;
            }
        );
    }

    setListBulleted(){
        this.setLineBlock(
            {
                type:           'listBulleted',
                forbiddenTypes: ['break', 'thematicBreak', 'image', 'imageReference', 'table', 'code']
            },
            (lines, modeAdd) => {
                for(var i = 0; i < lines.length; i++){
                    var n = this.normalizeLine(lines[i]);
                    var add = '';
                    if(modeAdd){
                        add = '- ';
                    }
                    lines[i] = n.whitespace + add + n.line;
                }
                return lines;
            }
        );
    }

    setListNumbered(){
        this.setLineBlock(
            {
                type:           'listNumbered',
                forbiddenTypes: ['break', 'thematicBreak', 'image', 'imageReference', 'table', 'code']
            },
            (lines, modeAdd) => {
                for(var i = 0; i < lines.length; i++){
                    var n = this.normalizeLine(lines[i]);
                    var add = '';
                    if(modeAdd){
                        var index = i + 1;
                        add = index.toString() + '. ';
                    }
                    lines[i] = n.whitespace + add + n.line;
                }
                return lines;
            }
        );
    }

    normalizeLine(line){
        var start = '';
        while(line.substr(0, 2) === '  '){
            start += '  ';
            line = line.substr(2);
        }
        line = line.replace(/^#{1,6}\s/g, ''); // heading
        line = line.replace(/^[-+*]\s/g, '');  // bulleted list
        line = line.replace(/^\d+\.\s/g, '');  // numbered list
        line = line.replace(/^>\s/g, '');      // blockquote
        return {
            whitespace: start,
            line:       line
        };
    }
    
    // functions
    
    setInlineBlock(type, forbiddenTypes, callback){
        var modeAdd = true;
        var s = this.props.store;
        var selectionStart = s.selectionStart;
        var selectionEnd = s.selectionEnd;
        for(var i = 0; i < forbiddenTypes.length; i++){
            if(s.selectedTypes.includes(forbiddenTypes[i])){
                console.log('Text-type change isn´t allowed here');
                return false;
            }
        }
        if(s.selectedTypes.includes(type)){
            modeAdd = false;
            for(var i = 0; i < s.selectedComponents.length; i++){
                if(s.selectedComponents[i].name === type){
                    selectionStart = s.selectedComponents[i].low;
                    selectionEnd   = s.selectedComponents[i].high;
                    break;
                }
            }
        }
        var split = s.splitString(s.content, selectionStart, selectionEnd);
        var line = callback(split.selected, modeAdd);
        s.onContentChange(split.pre + line + split.post, split.pre.length, split.pre.length + line.length);
    }

    setLineBlock(config, callback){
        var def = {
            type:           false,
            forbiddenTypes: [],
            completeType:   false
        };
        var c = {...def, ...config};

        var modeAdd = true;
        var s = this.props.store;
        for(var i = 0; i < c.forbiddenTypes.length; i++){
            if(s.selectedTypes.includes(c.forbiddenTypes[i])){
                console.log('Text-type change isn´t allowed here');
                return false;
            }
        }
        var split = s.splitLines(s.content, s.selectionStart, s.selectionEnd);
        if(c.type){
            var types = s.getTypesForRange(split.pre.length, split.pre.length + split.lines.join("\n").length + 1);
            if(types.includes(c.type)){
                modeAdd = false;
                if(c.completeType){
                    for(var i = 0; i < s.selectedComponents.length; i++){
                        if(s.selectedComponents[i].name === c.type){
                            split = s.splitLines(s.content, s.selectedComponents[i].low, s.selectedComponents[i].high);
                            break;
                        }
                    }
                }
            }
        }
        var lines = split.lines;
        lines = callback(lines, modeAdd);
        lines = lines.join("\n");
        var pre = split.pre + "\n";
        if(pre.trim() === ''){
            pre = '';
        }
        var post = "\n" + split.post;
        if(post.trim() === ''){
            post = '';
        }
        s.onContentChange(pre + lines + post, pre.length, pre.length + lines.length);
    }
    
    positionHijack(Component, store, type){
        return class extends React.Component {
            render() {
                const { sourcePosition, ...passThroughProps } = this.props;
                var typeName = type;
                if(type == 'heading'){
                    typeName = 'heading-' + passThroughProps.level.toString();
                }
                if(type == 'list'){
                    if(passThroughProps.ordered === true){
                        typeName = 'listNumbered';
                    } else {
                        typeName = 'listBulleted';
                    }
                }
                if(type == 'code'){
                    switch(passThroughProps.language){
                        case 'math':
                            typeName = 'math';
                            break;
                        case 'plotly':
                            typeName = 'plotly';
                            break;
                        default:
                            break;
                    }
                }
                store.addComponentToTree(sourcePosition.start.offset, sourcePosition.end.offset, typeName);
                return <Component {...passThroughProps} />;
            }
        }
    }

    render(){
        var s = this.props.store;
        var textarea = (
            <textarea class="form-control editor"
                onChange={this.onContentChange.bind(this)}
                onSelect={this.onSelectionChange.bind(this)}
                value={s.content}
                ref={(node) => {this.textarea = node}}
                onFocus={this.onFocus.bind(this)}
                onBlur={this.onBlur.bind(this)}
                onScroll={this.onScroll.bind(this)}
                >
            </textarea>
        );
        if(this.textarea){
            this.textarea.setSelectionRange(s.selectionStart, s.selectionEnd);
            this.textarea.focus();
        }
        var markdown = <ReactMarkdown source={s.content} rawSourcePos={true} transformLinkUri={null} renderers={{
            paragraph:  this.positionHijack(Paragraph, s, 'paragraph'),
            blockquote: this.positionHijack(Blockquote, s, 'blockquote'),
            break:      this.positionHijack(Break, s, 'break'),
            code:       this.positionHijack(Code, s, 'code'),
            inlineCode: this.positionHijack(InlineCode, s, 'inlineCode'),
            emphasis:   this.positionHijack(Emphasis, s, 'emphasis'),
            delete:     this.positionHijack(Delete, s, 'delete'),
            heading:    this.positionHijack(Heading, s, 'heading'),
            image:      this.positionHijack(Image, s, 'image'),
            table:      this.positionHijack(Table, s, 'table'),
            strong:     this.positionHijack(Strong, s, 'strong'),
            list:       this.positionHijack(List, s, 'list'),
            listItem:   this.positionHijack(ListItem, s, 'listItem'),
            link:       this.positionHijack(Link, s, 'link'),
        }}/>;
        var textTypes = {
            'paragraph': 'Paragraph',
            'heading-1': 'Header 1',
            'heading-2': 'Header 2',
            'heading-3': 'Header 3',
            'heading-4': 'Header 4',
            'heading-5': 'Header 5',
            'heading-6': 'Header 6'
        };
        return (
            <div class="row editor">
                <div class="col-6">
                    <div class="row mb-2">
                        <div class="col-12">
                            <div class="btn-group mr-2" role="file" aria-label="Basic example">
                                <button type="button" class="btn" onClick={this.saveFile.bind(this)}><Icon>save_alt</Icon></button>
                            </div>
                            <div class="btn-group mr-2" role="text" aria-label="Basic example">
                                <ButtonDropdown isOpen={this.state.textDropdown} toggle={this.onTextDropdownToggle.bind(this)}>
                                    <DropdownToggle caret color="none">
                                        {textTypes[s.selectedTextType]}
                                    </DropdownToggle>
                                    <DropdownMenu>
                                        <DropdownItem active={s.isTypeSelected('paragraph')} onClick={this.setParagraph.bind(this)}><p>Paragraph</p></DropdownItem>
                                        <DropdownItem active={s.isTypeSelected('heading-1')} onClick={this.setHeader1.bind(this)}><h1>Header 1</h1></DropdownItem>
                                        <DropdownItem active={s.isTypeSelected('heading-2')} onClick={this.setHeader2.bind(this)}><h2>Header 2</h2></DropdownItem>
                                        <DropdownItem active={s.isTypeSelected('heading-3')} onClick={this.setHeader3.bind(this)}><h3>Header 3</h3></DropdownItem>
                                        <DropdownItem active={s.isTypeSelected('heading-4')} onClick={this.setHeader4.bind(this)}><h4>Header 4</h4></DropdownItem>
                                        <DropdownItem active={s.isTypeSelected('heading-5')} onClick={this.setHeader5.bind(this)}><h5>Header 5</h5></DropdownItem>
                                        <DropdownItem active={s.isTypeSelected('heading-6')} onClick={this.setHeader6.bind(this)}><h6>Header 6</h6></DropdownItem>
                                    </DropdownMenu>
                                </ButtonDropdown>
                                <button type="button" class={"btn " + (s.isTypeSelected('strong') ? 'btn-primary' : '')} onClick={this.setStrong.bind(this)}><Icon>format_bold</Icon></button>
                                <button type="button" class={"btn " + (s.isTypeSelected('emphasis') ? 'btn-primary' : '')} onClick={this.setEmphasis.bind(this)}><Icon>format_italic</Icon></button>
                                <button type="button" class={"btn " + (s.isTypeSelected('delete') ? 'btn-primary' : '')} onClick={this.setDelete.bind(this)}><Icon>format_strikethrough</Icon></button>
                                <button type="button" class={"btn " + (s.isTypeSelected('listBulleted') ? 'btn-primary' : '')} onClick={this.setListBulleted.bind(this)}><Icon>format_list_bulleted</Icon></button>
                                <button type="button" class={"btn " + (s.isTypeSelected('listNumbered') ? 'btn-primary' : '')} onClick={this.setListNumbered.bind(this)}><Icon>format_list_numbered</Icon></button>
                                <button type="button" class={"btn " + (s.isTypeSelected('code') ? 'btn-primary' : '')} onClick={this.setCode.bind(this)}><Icon>code</Icon></button>
                                <button type="button" class={"btn " + (s.isTypeSelected('blockquote') ? 'btn-primary' : '')} onClick={this.setBlockquote.bind(this)}><Icon>format_quote</Icon></button>
                            </div>
                            <div class="btn-group" role="components" aria-label="Basic example">
                                <button type="button" class={"btn " + (s.isTypeSelected('link') ? 'btn-primary' : '')} onClick={this.setLink.bind(this)}><Icon>insert_link</Icon></button>
                                <button type="button" class={"btn " + (s.isTypeSelected('image') ? 'btn-primary' : '')} onClick={this.setImage.bind(this)}><Icon>insert_photo</Icon></button>
                                <button type="button" class={"btn " + (s.isTypeSelected('plotly') ? 'btn-primary' : '')} onClick={this.setPlotly.bind(this)}><Icon>insert_chart</Icon></button>
                                <button type="button" class={"btn " + (s.isTypeSelected('math') ? 'btn-primary' : '')} onClick={this.setMath.bind(this)}><Icon>functions</Icon></button>
                            </div>
                        </div>
                    </div>
                    <div class="row">
                        <div class="col-12">
                            {textarea}
                        </div>
                    </div>
                </div>
                <div class="col-6">
                    {markdown}
                </div>
            </div>
        );
    }
}